@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2025, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -32,26 +32,42 @@ namespace mfem
 
 class BilinearForm;
 
-/// Abstract base class for an iterative solver monitor
-class IterativeSolverMonitor
+/// Abstract base class for an iterative solver controller
+class IterativeSolverController
 {
 protected:
-   /// The last IterativeSolver to which this monitor was attached.
+   /// The last IterativeSolver to which this controller was attached.
    const class IterativeSolver *iter_solver;
 
+   /// In MonitorResidual or MonitorSolution, this member variable can be set
+   /// to true to indicate early convergence.
+   bool converged = false;
+
 public:
-   IterativeSolverMonitor() : iter_solver(nullptr) {}
+   IterativeSolverController() : iter_solver(nullptr) {}
 
-   virtual ~IterativeSolverMonitor() {}
+   virtual ~IterativeSolverController() {}
 
-   /// Monitor the residual vector r
-   virtual void MonitorResidual(int it, double norm, const Vector &r,
+   /// Has the solver converged?
+   ///
+   /// Can be used if convergence is detected in the controller (before reaching
+   /// the relative or absolute tolerance of the IterativeSolver).
+   bool HasConverged() { return converged; }
+
+   /// Reset the controller to its initial state.
+   ///
+   /// This function is called by the IterativeSolver::Mult()
+   /// method on the first iteration.
+   virtual void Reset() { converged = false; }
+
+   /// Monitor the solution vector r
+   virtual void MonitorResidual(int it, real_t norm, const Vector &r,
                                 bool final)
    {
    }
 
    /// Monitor the solution vector x
-   virtual void MonitorSolution(int it, double norm, const Vector &x,
+   virtual void MonitorSolution(int it, real_t norm, const Vector &x,
                                 bool final)
    {
    }
@@ -62,6 +78,9 @@ public:
    { iter_solver = &solver; }
 };
 
+/// Keeping the alias for backward compatibility
+using IterativeSolverMonitor = IterativeSolverController;
+
 /// Abstract base class for iterative solver
 class IterativeSolver : public Solver
 {
@@ -106,7 +125,8 @@ public:
       PrintLevel &Iterations() { iterations=true; return *this; }
       PrintLevel &FirstAndLast() { first_and_last=true; return *this; }
       PrintLevel &Summary() { summary=true; return *this; }
-      PrintLevel &All() { return Warnings().Errors().Iterations().FirstAndLast().Summary(); }
+      PrintLevel &All()
+      { return Warnings().Errors().Iterations().FirstAndLast().Summary(); }
       ///@}
    };
 
@@ -151,10 +171,10 @@ protected:
    int max_iter;
 
    /// Relative tolerance.
-   double rel_tol;
+   real_t rel_tol;
 
    /// Absolute tolerance.
-   double abs_tol;
+   real_t abs_tol;
 
    ///@}
 
@@ -164,13 +184,23 @@ protected:
 
    mutable int final_iter = -1;
    mutable bool converged = false;
-   mutable double initial_norm = -1.0, final_norm = -1.0;
+   mutable real_t initial_norm = -1.0, final_norm = -1.0;
 
    ///@}
 
-   double Dot(const Vector &x, const Vector &y) const;
-   double Norm(const Vector &x) const { return sqrt(Dot(x, x)); }
-   void Monitor(int it, double norm, const Vector& r, const Vector& x,
+
+   /** @brief Return the standard (l2, i.e., Euclidean) inner product of
+       @a x and @a y
+       @details Overriding this method in a derived class enables a
+       custom inner product.
+      */
+   virtual real_t Dot(const Vector &x, const Vector &y) const;
+
+   /// Return the inner product norm of @a x, using the inner product defined by Dot()
+   real_t Norm(const Vector &x) const { return sqrt(Dot(x, x)); }
+
+   /// Monitor both the residual @a r and the solution @a x
+   bool Monitor(int it, real_t norm, const Vector& r, const Vector& x,
                 bool final=false) const;
 
 public:
@@ -186,18 +216,18 @@ public:
        @details While the convergence criterion is solver specific, most of the
        provided iterative solvers use one of the following criteria
 
-       \f$ ||r||_X \leq tol_{rel}||r_0||_X \f$,
+       $ ||r||_X \leq tol_{rel}||r_0||_X $,
 
-       \f$ ||r||_X \leq tol_{abs} \f$,
+       $ ||r||_X \leq tol_{abs} $,
 
-       \f$ ||r||_X \leq \max\{ tol_{abs}, tol_{rel} ||r_0||_X \} \f$,
+       $ ||r||_X \leq \max\{ tol_{abs}, tol_{rel} ||r_0||_X \} $,
 
        where X denotes the space in which the norm is measured. The choice of
        X depends on the specific iterative solver.
       */
    ///@{
-   void SetRelTol(double rtol) { rel_tol = rtol; }
-   void SetAbsTol(double atol) { abs_tol = atol; }
+   void SetRelTol(real_t rtol) { rel_tol = rtol; }
+   void SetAbsTol(real_t atol) { abs_tol = atol; }
    void SetMaxIter(int max_it) { max_iter = max_it; }
    ///@}
 
@@ -255,20 +285,20 @@ public:
    /// This function returns the norm of the residual (or preconditioned
    /// residual, depending on the solver), computed before the start of the
    /// iteration.
-   double GetInitialNorm() const { return initial_norm; }
+   real_t GetInitialNorm() const { return initial_norm; }
    /// @brief Returns the final residual norm after termination of the solver
    /// during the last call to Mult().
    ///
    /// This function returns the norm of the residual (or preconditioned
    /// residual, depending on the solver), corresponding to the returned
    /// solution.
-   double GetFinalNorm() const { return final_norm; }
+   real_t GetFinalNorm() const { return final_norm; }
    /// @brief Returns the final residual norm after termination of the solver
    /// during the last call to Mult(), divided by the initial residual norm.
    /// Returns -1 if one of these norms is left undefined by the solver.
    ///
    /// @sa GetFinalNorm(), GetInitialNorm()
-   double GetFinalRelNorm() const
+   real_t GetFinalRelNorm() const
    {
       if (final_norm < 0.0 || initial_norm < 0.0) { return -1.0; }
       return final_norm / initial_norm;
@@ -280,7 +310,7 @@ public:
    virtual void SetPreconditioner(Solver &pr);
 
    /// Also calls SetOperator for the preconditioner if there is one
-   virtual void SetOperator(const Operator &op) override;
+   void SetOperator(const Operator &op) override;
 
    /// Set the iterative solver monitor
    void SetMonitor(IterativeSolverMonitor &m)
@@ -305,7 +335,7 @@ public:
    /** @brief Default constructor: the diagonal will be computed by subsequent
        calls to SetOperator() using the Operator method AssembleDiagonal. */
    /** In this case the array of essential tdofs will be empty. */
-   OperatorJacobiSmoother(const double damping=1.0);
+   OperatorJacobiSmoother(const real_t damping=1.0);
 
    /** Setup a Jacobi smoother with the diagonal of @a a obtained by calling
        a.AssembleDiagonal(). It is assumed that the underlying operator acts as
@@ -318,7 +348,7 @@ public:
        original behavior of this class. */
    OperatorJacobiSmoother(const BilinearForm &a,
                           const Array<int> &ess_tdof_list,
-                          const double damping=1.0);
+                          const real_t damping=1.0);
 
    /** Application is by the *inverse* of the given vector. It is assumed that
        the underlying operator acts as the identity on entries in ess_tdof_list,
@@ -331,14 +361,18 @@ public:
        original behavior of this class. */
    OperatorJacobiSmoother(const Vector &d,
                           const Array<int> &ess_tdof_list,
-                          const double damping=1.0);
+                          const real_t damping=1.0);
 
    ~OperatorJacobiSmoother() {}
 
    /// Replace diagonal entries with their absolute values.
    void SetPositiveDiagonal(bool pos_diag = true) { use_abs_diag = pos_diag; }
 
+   /// Approach the solution of the linear system by applying Jacobi smoothing.
    void Mult(const Vector &x, Vector &y) const;
+
+   /** @brief Approach the solution of the transposed linear system by applying
+       Jacobi smoothing. */
    void MultTranspose(const Vector &x, Vector &y) const { Mult(x, y); }
 
    /** @brief Recompute the diagonal using the method AssembleDiagonal of the
@@ -356,7 +390,7 @@ public:
 
 private:
    Vector dinv;
-   const double damping;
+   const real_t damping;
    const Array<int> *ess_tdof_list; // not owned; may be NULL
    mutable Vector residual;
    /// Uses absolute values of the diagonal entries.
@@ -387,13 +421,13 @@ public:
        max_eig_estimate. */
    OperatorChebyshevSmoother(const Operator &oper_, const Vector &d,
                              const Array<int>& ess_tdof_list,
-                             int order, double max_eig_estimate);
+                             int order, real_t max_eig_estimate);
 
    /// Deprecated: see pass-by-reference version above
    MFEM_DEPRECATED
    OperatorChebyshevSmoother(const Operator* oper_, const Vector &d,
                              const Array<int>& ess_tdof_list,
-                             int order, double max_eig_estimate);
+                             int order, real_t max_eig_estimate);
 
    /** Application is by *inverse* of the given vector. It is assumed the
        underlying operator acts as the identity on entries in ess_tdof_list,
@@ -407,7 +441,8 @@ public:
                              const Array<int>& ess_tdof_list,
                              int order, MPI_Comm comm = MPI_COMM_NULL,
                              int power_iterations = 10,
-                             double power_tolerance = 1e-8);
+                             real_t power_tolerance = 1e-8,
+                             int power_seed = 12345);
 
    /// Deprecated: see pass-by-reference version above
    MFEM_DEPRECATED
@@ -415,25 +450,30 @@ public:
                              const Array<int>& ess_tdof_list,
                              int order, MPI_Comm comm = MPI_COMM_NULL,
                              int power_iterations = 10,
-                             double power_tolerance = 1e-8);
+                             real_t power_tolerance = 1e-8);
 #else
    OperatorChebyshevSmoother(const Operator &oper_, const Vector &d,
                              const Array<int>& ess_tdof_list,
                              int order, int power_iterations = 10,
-                             double power_tolerance = 1e-8);
+                             real_t power_tolerance = 1e-8,
+                             int power_seed = 12345);
 
    /// Deprecated: see pass-by-reference version above
    MFEM_DEPRECATED
    OperatorChebyshevSmoother(const Operator* oper_, const Vector &d,
                              const Array<int>& ess_tdof_list,
                              int order, int power_iterations = 10,
-                             double power_tolerance = 1e-8);
+                             real_t power_tolerance = 1e-8);
 #endif
 
    ~OperatorChebyshevSmoother() {}
 
-   void Mult(const Vector&x, Vector &y) const;
+   /** @brief Approach the solution of the linear system by applying Chebyshev
+       smoothing. */
+   void Mult(const Vector &x, Vector &y) const;
 
+   /** @brief Approach the solution of the transposed linear system by applying
+       Chebyshev smoothing. */
    void MultTranspose(const Vector &x, Vector &y) const { Mult(x, y); }
 
    void SetOperator(const Operator &op_)
@@ -445,11 +485,11 @@ public:
 
 private:
    const int order;
-   double max_eig_estimate;
+   real_t max_eig_estimate;
    const int N;
    Vector dinv;
    const Vector &diag;
-   Array<double> coeffs;
+   Array<real_t> coeffs;
    const Array<int>& ess_tdof_list;
    mutable Vector residual;
    mutable Vector helperVector;
@@ -472,21 +512,25 @@ public:
    SLISolver(MPI_Comm comm_) : IterativeSolver(comm_) { }
 #endif
 
-   virtual void SetOperator(const Operator &op)
+   void SetOperator(const Operator &op) override
    { IterativeSolver::SetOperator(op); UpdateVectors(); }
 
-   virtual void Mult(const Vector &b, Vector &x) const;
+   /// Iterative solution of the linear system using Stationary Linear Iteration
+   /** When using iterative mode (see Solver::iterative_mode), the case of zero
+       r.h.s., @a b = 0, can be optimized by calling this method with empty
+       @a b, i.e. b.Size() == 0. */
+   void Mult(const Vector &b, Vector &x) const override;
 };
 
 /// Stationary linear iteration. (tolerances are squared)
 void SLI(const Operator &A, const Vector &b, Vector &x,
          int print_iter = 0, int max_num_iter = 1000,
-         double RTOLERANCE = 1e-12, double ATOLERANCE = 1e-24);
+         real_t RTOLERANCE = 1e-12, real_t ATOLERANCE = 1e-24);
 
 /// Preconditioned stationary linear iteration. (tolerances are squared)
 void SLI(const Operator &A, Solver &B, const Vector &b, Vector &x,
          int print_iter = 0, int max_num_iter = 1000,
-         double RTOLERANCE = 1e-12, double ATOLERANCE = 1e-24);
+         real_t RTOLERANCE = 1e-12, real_t ATOLERANCE = 1e-24);
 
 
 /// Conjugate gradient method
@@ -504,21 +548,23 @@ public:
    CGSolver(MPI_Comm comm_) : IterativeSolver(comm_) { }
 #endif
 
-   virtual void SetOperator(const Operator &op)
+   void SetOperator(const Operator &op) override
    { IterativeSolver::SetOperator(op); UpdateVectors(); }
 
-   virtual void Mult(const Vector &b, Vector &x) const;
+   /** @brief Iterative solution of the linear system using the Conjugate
+       Gradient method. */
+   void Mult(const Vector &b, Vector &x) const override;
 };
 
 /// Conjugate gradient method. (tolerances are squared)
 void CG(const Operator &A, const Vector &b, Vector &x,
         int print_iter = 0, int max_num_iter = 1000,
-        double RTOLERANCE = 1e-12, double ATOLERANCE = 1e-24);
+        real_t RTOLERANCE = 1e-12, real_t ATOLERANCE = 1e-24);
 
 /// Preconditioned conjugate gradient method. (tolerances are squared)
 void PCG(const Operator &A, Solver &B, const Vector &b, Vector &x,
          int print_iter = 0, int max_num_iter = 1000,
-         double RTOLERANCE = 1e-12, double ATOLERANCE = 1e-24);
+         real_t RTOLERANCE = 1e-12, real_t ATOLERANCE = 1e-24);
 
 
 /// GMRES method
@@ -537,7 +583,8 @@ public:
    /// Set the number of iteration to perform between restarts, default is 50.
    void SetKDim(int dim) { m = dim; }
 
-   virtual void Mult(const Vector &b, Vector &x) const;
+   /// Iterative solution of the linear system using the GMRES method
+   void Mult(const Vector &b, Vector &x) const override;
 };
 
 /// FGMRES method
@@ -555,17 +602,18 @@ public:
 
    void SetKDim(int dim) { m = dim; }
 
-   virtual void Mult(const Vector &b, Vector &x) const;
+   /// Iterative solution of the linear system using the FGMRES method.
+   void Mult(const Vector &b, Vector &x) const override;
 };
 
 /// GMRES method. (tolerances are squared)
 int GMRES(const Operator &A, Vector &x, const Vector &b, Solver &M,
-          int &max_iter, int m, double &tol, double atol, int printit);
+          int &max_iter, int m, real_t &tol, real_t atol, int printit);
 
 /// GMRES method. (tolerances are squared)
 void GMRES(const Operator &A, Solver &B, const Vector &b, Vector &x,
            int print_iter = 0, int max_num_iter = 1000, int m = 50,
-           double rtol = 1e-12, double atol = 1e-24);
+           real_t rtol = 1e-12, real_t atol = 1e-24);
 
 
 /// BiCGSTAB method
@@ -583,20 +631,21 @@ public:
    BiCGSTABSolver(MPI_Comm comm_) : IterativeSolver(comm_) { }
 #endif
 
-   virtual void SetOperator(const Operator &op)
+   void SetOperator(const Operator &op) override
    { IterativeSolver::SetOperator(op); UpdateVectors(); }
 
-   virtual void Mult(const Vector &b, Vector &x) const;
+   /// Iterative solution of the linear system using the BiCGSTAB method
+   void Mult(const Vector &b, Vector &x) const override;
 };
 
 /// BiCGSTAB method. (tolerances are squared)
 int BiCGSTAB(const Operator &A, Vector &x, const Vector &b, Solver &M,
-             int &max_iter, double &tol, double atol, int printit);
+             int &max_iter, real_t &tol, real_t atol, int printit);
 
 /// BiCGSTAB method. (tolerances are squared)
 void BiCGSTAB(const Operator &A, Solver &B, const Vector &b, Vector &x,
               int print_iter = 0, int max_num_iter = 1000,
-              double rtol = 1e-12, double atol = 1e-24);
+              real_t rtol = 1e-12, real_t atol = 1e-24);
 
 
 /// MINRES method
@@ -613,25 +662,26 @@ public:
    MINRESSolver(MPI_Comm comm_) : IterativeSolver(comm_) { }
 #endif
 
-   virtual void SetPreconditioner(Solver &pr)
+   void SetPreconditioner(Solver &pr) override
    {
       IterativeSolver::SetPreconditioner(pr);
       if (oper) { u1.SetSize(width); }
    }
 
-   virtual void SetOperator(const Operator &op);
+   void SetOperator(const Operator &op) override;
 
-   virtual void Mult(const Vector &b, Vector &x) const;
+   /// Iterative solution of the linear system using the MINRES method
+   void Mult(const Vector &b, Vector &x) const override;
 };
 
 /// MINRES method without preconditioner. (tolerances are squared)
 void MINRES(const Operator &A, const Vector &b, Vector &x, int print_it = 0,
-            int max_it = 1000, double rtol = 1e-12, double atol = 1e-24);
+            int max_it = 1000, real_t rtol = 1e-12, real_t atol = 1e-24);
 
 /// MINRES method with preconditioner. (tolerances are squared)
 void MINRES(const Operator &A, Solver &B, const Vector &b, Vector &x,
             int print_it = 0, int max_it = 1000,
-            double rtol = 1e-12, double atol = 1e-24);
+            real_t rtol = 1e-12, real_t atol = 1e-24);
 
 
 /// Newton's method for solving F(x)=b for a given operator F.
@@ -649,32 +699,32 @@ protected:
    // Method to determine rtol, 0 means the adaptive algorithm is deactivated.
    int lin_rtol_type = 0;
    // rtol to use in first iteration
-   double lin_rtol0;
+   real_t lin_rtol0;
    // Maximum rtol
-   double lin_rtol_max;
+   real_t lin_rtol_max;
    // Function norm ||F(x)|| of the previous iterate
-   mutable double fnorm_last = 0.0;
+   mutable real_t fnorm_last = 0.0;
    // Linear residual norm of the previous iterate
-   mutable double lnorm_last = 0.0;
+   mutable real_t lnorm_last = 0.0;
    // Forcing term (linear residual rtol) from the previous iterate
-   mutable double eta_last = 0.0;
+   mutable real_t eta_last = 0.0;
    // Eisenstat-Walker factor gamma
-   double gamma;
+   real_t gamma;
    // Eisenstat-Walker factor alpha
-   double alpha;
+   real_t alpha;
 
    /** @brief Method for the adaptive linear solver rtol invoked before the
        linear solve. */
    void AdaptiveLinRtolPreSolve(const Vector &x,
                                 const int it,
-                                const double fnorm) const;
+                                const real_t fnorm) const;
 
    /** @brief Method for the adaptive linear solver rtol invoked after the
        linear solve. */
    void AdaptiveLinRtolPostSolve(const Vector &x,
                                  const Vector &b,
                                  const int it,
-                                 const double fnorm) const;
+                                 const real_t fnorm) const;
 
 public:
    NewtonSolver() { }
@@ -682,7 +732,7 @@ public:
 #ifdef MFEM_USE_MPI
    NewtonSolver(MPI_Comm comm_) : IterativeSolver(comm_) { }
 #endif
-   virtual void SetOperator(const Operator &op);
+   void SetOperator(const Operator &op) override;
 
    /// Set the linear solver for inverting the Jacobian.
    /** This method is equivalent to calling SetPreconditioner(). */
@@ -690,13 +740,13 @@ public:
 
    /// Solve the nonlinear system with right-hand side @a b.
    /** If `b.Size() != Height()`, then @a b is assumed to be zero. */
-   virtual void Mult(const Vector &b, Vector &x) const;
+   void Mult(const Vector &b, Vector &x) const override;
 
    /** @brief This method can be overloaded in derived classes to implement line
        search algorithms. */
    /** The base class implementation (NewtonSolver) simply returns 1. A return
        value of 0 indicates a failure, interrupting the Newton iteration. */
-   virtual double ComputeScalingFactor(const Vector &x, const Vector &b) const
+   virtual real_t ComputeScalingFactor(const Vector &x, const Vector &b) const
    { return 1.0; }
 
    /** @brief This method can be overloaded in derived classes to perform
@@ -715,10 +765,10 @@ public:
     in an inexact Newton method."
     */
    void SetAdaptiveLinRtol(const int type = 2,
-                           const double rtol0 = 0.5,
-                           const double rtol_max = 0.9,
-                           const double alpha = 0.5 * (1.0 + sqrt(5.0)),
-                           const double gamma = 1.0);
+                           const real_t rtol0 = 0.5,
+                           const real_t rtol_max = 0.9,
+                           const real_t alpha = 0.5 * (1.0 + sqrt(5.0)),
+                           const real_t gamma = 1.0);
 };
 
 /** L-BFGS method for solving F(x)=b for a given operator F, by minimizing
@@ -759,7 +809,7 @@ public:
    LBFGSSolver(MPI_Comm comm_) : NewtonSolver(comm_) { }
 #endif
 
-   virtual void SetOperator(const Operator &op)
+   void SetOperator(const Operator &op) override
    {
       NewtonSolver::SetOperator(op);
       InitializeStorageVectors();
@@ -773,11 +823,11 @@ public:
 
    /// Solve the nonlinear system with right-hand side @a b.
    /** If `b.Size() != Height()`, then @a b is assumed to be zero. */
-   virtual void Mult(const Vector &b, Vector &x) const;
+   void Mult(const Vector &b, Vector &x) const override;
 
-   virtual void SetPreconditioner(Solver &pr)
+   void SetPreconditioner(Solver &pr) override
    { MFEM_WARNING("L-BFGS won't use the given preconditioner."); }
-   virtual void SetSolver(Solver &solver)
+   void SetSolver(Solver &solver) override
    { MFEM_WARNING("L-BFGS won't use the given solver."); }
 
    virtual ~LBFGSSolver() { DeleteStorageVectors(); }
@@ -790,9 +840,12 @@ public:
     cf(=0.4) is a desired convergence factor. */
 int aGMRES(const Operator &A, Vector &x, const Vector &b,
            const Operator &M, int &max_iter,
-           int m_max, int m_min, int m_step, double cf,
-           double &tol, double &atol, int printit);
+           int m_max, int m_min, int m_step, real_t cf,
+           real_t &tol, real_t &atol, int printit);
 
+#ifdef MFEM_USE_HIOP
+class HiopOptimizationProblem;
+#endif
 
 /** Defines operators and constraints for the following optimization problem:
  *
@@ -812,11 +865,25 @@ int aGMRES(const Operator &A, Vector &x, const Vector &b,
  *  the operators are expected to be defined for tdof vectors. */
 class OptimizationProblem
 {
+#ifdef MFEM_USE_HIOP
+   friend class HiopOptimizationProblem;
+#endif
+
+private:
+   /// See NewX().
+   mutable bool new_x = true;
+
 protected:
    /// Not owned, some can remain unused (NULL).
    const Operator *C, *D;
    const Vector *c_e, *d_lo, *d_hi, *x_lo, *x_hi;
 
+   /// Implementations of CalcObjective() and CalcObjectiveGrad() can use this
+   /// method to check if the argument Vector x has been changed after the last
+   /// call to CalcObjective() or CalcObjectiveGrad().
+   /// The result is on by default, and gets set by the OptimizationSolver.
+   bool NewX() const { return new_x; }
+
 public:
    const int input_size;
 
@@ -824,7 +891,7 @@ public:
    OptimizationProblem(int insize, const Operator *C_, const Operator *D_);
 
    /// Objective F(x). In parallel, the result should be reduced over tasks.
-   virtual double CalcObjective(const Vector &x) const = 0;
+   virtual real_t CalcObjective(const Vector &x) const = 0;
    /// The result grad is expected to enter with the correct size.
    virtual void CalcObjectiveGrad(const Vector &x, Vector &grad) const
    { MFEM_ABORT("The objective gradient is not implemented."); }
@@ -855,18 +922,18 @@ public:
 #ifdef MFEM_USE_MPI
    OptimizationSolver(MPI_Comm comm_): IterativeSolver(comm_), problem(NULL) { }
 #endif
-   virtual ~OptimizationSolver() { }
+   ~OptimizationSolver() override { }
 
    /** This function is virtual as solvers might need to perform some initial
     *  actions (e.g. validation) with the OptimizationProblem. */
    virtual void SetOptimizationProblem(const OptimizationProblem &prob)
    { problem = &prob; }
 
-   virtual void Mult(const Vector &xt, Vector &x) const = 0;
+   void Mult(const Vector &xt, Vector &x) const override = 0;
 
-   virtual void SetPreconditioner(Solver &pr)
+   void SetPreconditioner(Solver &pr) override
    { MFEM_ABORT("Not meaningful for this solver."); }
-   virtual void SetOperator(const Operator &op)
+   void SetOperator(const Operator &op) override
    { MFEM_ABORT("Not meaningful for this solver."); }
 };
 
@@ -881,10 +948,10 @@ class SLBQPOptimizer : public OptimizationSolver
 {
 protected:
    Vector lo, hi, w;
-   double a;
+   real_t a;
 
    /// Solve QP at fixed lambda
-   inline double solve(double l, const Vector &xt, Vector &x, int &nclip) const
+   inline real_t solve(real_t l, const Vector &xt, Vector &x, int &nclip) const
    {
       add(xt, l, w, x);
       if (problem == NULL) { x.median(lo,hi); }
@@ -905,7 +972,7 @@ protected:
       }
    }
 
-   inline void print_iteration(int it, double r, double l) const;
+   inline void print_iteration(int it, real_t r, real_t l) const;
 
 public:
    SLBQPOptimizer() { }
@@ -917,14 +984,14 @@ public:
    /** Setting an OptimizationProblem will overwrite the Vectors given by
     *  SetBounds and SetLinearConstraint. The objective function remains
     *  unchanged. */
-   virtual void SetOptimizationProblem(const OptimizationProblem &prob);
+   void SetOptimizationProblem(const OptimizationProblem &prob) override;
 
    void SetBounds(const Vector &lo_, const Vector &hi_);
-   void SetLinearConstraint(const Vector &w_, double a_);
+   void SetLinearConstraint(const Vector &w_, real_t a_);
 
    /** We let the target values play the role of the initial vector xt, from
     *  which the operator generates the optimal vector x. */
-   virtual void Mult(const Vector &xt, Vector &x) const;
+   void Mult(const Vector &xt, Vector &x) const override;
 };
 
 /** Block ILU solver:
@@ -992,7 +1059,7 @@ public:
    /** Get the data array for the block CSR representation of the factorization.
     *  Similar to SparseMatrix::GetData(). Mostly used for testing.
     */
-   double *GetBlockData() { return AB.Data(); }
+   real_t *GetBlockData() { return AB.Data(); }
 
 private:
    /// Set up the block CSR structure corresponding to a sparse matrix @a A
@@ -1042,7 +1109,7 @@ public:
    ResidualBCMonitor(const Array<int> &ess_dofs_list_)
       : ess_dofs_list(&ess_dofs_list_) { }
 
-   void MonitorResidual(int it, double norm, const Vector &r,
+   void MonitorResidual(int it, real_t norm, const Vector &r,
                         bool final) override;
 };
 
@@ -1061,8 +1128,8 @@ protected:
    void Init();
 
 public:
-   double Control[UMFPACK_CONTROL];
-   mutable double Info[UMFPACK_INFO];
+   real_t Control[UMFPACK_CONTROL];
+   mutable real_t Info[UMFPACK_INFO];
 
    /** @brief For larger matrices, if the solver fails, set the parameter @a
        use_long_ints_ = true. */
@@ -1079,13 +1146,16 @@ public:
        The factorization uses the parameters set in the #Control data member.
        @note This method calls SparseMatrix::SortColumnIndices() with @a op,
        modifying the matrix if the column indices are not already sorted. */
-   virtual void SetOperator(const Operator &op);
+   void SetOperator(const Operator &op) override;
 
    /// Set the print level field in the #Control data member.
    void SetPrintLevel(int print_lvl) { Control[UMFPACK_PRL] = print_lvl; }
 
-   virtual void Mult(const Vector &b, Vector &x) const;
-   virtual void MultTranspose(const Vector &b, Vector &x) const;
+   /// Direct solution of the linear system using UMFPACK
+   void Mult(const Vector &b, Vector &x) const override;
+
+   /// Direct solution of the transposed linear system using UMFPACK
+   void MultTranspose(const Vector &b, Vector &x) const override;
 
    virtual ~UMFPackSolver();
 };
@@ -1109,10 +1179,13 @@ public:
    { Init(); SetOperator(A); }
 
    // Works on sparse matrices only; calls SparseMatrix::SortColumnIndices().
-   virtual void SetOperator(const Operator &op);
+   void SetOperator(const Operator &op) override;
 
-   virtual void Mult(const Vector &b, Vector &x) const;
-   virtual void MultTranspose(const Vector &b, Vector &x) const;
+   /// Direct solution of the linear system using KLU
+   void Mult(const Vector &b, Vector &x) const override;
+
+   /// Direct solution of the transposed linear system using KLU
+   void MultTranspose(const Vector &b, Vector &x) const override;
 
    virtual ~KLUSolver();
 
@@ -1133,8 +1206,10 @@ public:
    /// block_dof is a boolean matrix, block_dof(i, j) = 1 if j-th dof belongs to
    /// i-th block, block_dof(i, j) = 0 otherwise.
    DirectSubBlockSolver(const SparseMatrix& A, const SparseMatrix& block_dof);
-   virtual void Mult(const Vector &x, Vector &y) const;
-   virtual void SetOperator(const Operator &op) { }
+
+   /// Direct solution of the block diagonal linear system
+   void Mult(const Vector &x, Vector &y) const override;
+   void SetOperator(const Operator &op) override { }
 };
 
 /// Solver S such that I - A * S = (I - A * S1) * (I - A * S0).
@@ -1148,9 +1223,13 @@ public:
    ProductSolver(Operator* A_, Solver* S0_, Solver* S1_,
                  bool ownA, bool ownS0, bool ownS1)
       : Solver(A_->NumRows()), A(A_, ownA), S0(S0_, ownS0), S1(S1_, ownS1) { }
-   virtual void Mult(const Vector &x, Vector &y) const;
-   virtual void MultTranspose(const Vector &x, Vector &y) const;
-   virtual void SetOperator(const Operator &op) { }
+
+   /// Solution of the linear system using a product of subsolvers
+   void Mult(const Vector &x, Vector &y) const override;
+
+   /// Solution of the transposed linear system using a product of subsolvers
+   void MultTranspose(const Vector &x, Vector &y) const override;
+   void SetOperator(const Operator &op) override { }
 };
 
 /// Solver wrapper which orthogonalizes the input and output vector
@@ -1187,14 +1266,14 @@ public:
    /** The Operator @a op is simply forwarded to the solver object given by
        SetSolver() which needs to be called before this method. Calling this
        method is optional when the solver already has an associated Operator. */
-   virtual void SetOperator(const Operator &op);
+   void SetOperator(const Operator &op) override;
 
    /** @brief Perform the action of the OrthoSolver: P * solver * P where P is
        the projection to the subspace of vectors with zero sum. */
    /** @note The projection P can be written as P = I - 1 1^T / (1^T 1) where
        I is the identity matrix and 1 is the column-vector with all components
        equal to 1. */
-   void Mult(const Vector &b, Vector &x) const;
+   void Mult(const Vector &b, Vector &x) const override;
 
 private:
    Solver *solver = nullptr;
@@ -1220,14 +1299,141 @@ class AuxSpaceSmoother : public Solver
 public:
    AuxSpaceSmoother(const HypreParMatrix &op, HypreParMatrix *aux_map,
                     bool op_is_symmetric = true, bool own_aux_map = false);
-   virtual void Mult(const Vector &x, Vector &y) const { Mult(x, y, false); }
-   virtual void MultTranspose(const Vector &x, Vector &y) const { Mult(x, y, true); }
-   virtual void SetOperator(const Operator &op) { }
+   void Mult(const Vector &x, Vector &y) const override { Mult(x, y, false); }
+   void MultTranspose(const Vector &x, Vector &y) const override
+   { Mult(x, y, true); }
+   void SetOperator(const Operator &op) override { }
    HypreSmoother& GetSmoother() { return *aux_smoother_.As<HypreSmoother>(); }
    using Operator::Mult;
 };
 #endif // MFEM_USE_MPI
 
+#ifdef MFEM_USE_LAPACK
+/** Non-negative least squares (NNLS) solver class, for computing a vector
+    with non-negative entries approximately satisfying an under-determined
+    linear system. */
+class NNLSSolver : public Solver
+{
+public:
+   NNLSSolver();
+
+   ~NNLSSolver() { }
+
+   /// The operator must be a DenseMatrix.
+   void SetOperator(const Operator &op) override;
+
+   /** @brief Compute the non-negative least squares solution to the
+       underdetermined system. */
+   void Mult(const Vector &w, Vector &sol) const override;
+
+   /** @brief
+       Set verbosity. If set to 0: print nothing; if 1: just print results;
+       if 2: print short update on every iteration; if 3: print longer update
+       each iteration.
+     */
+   void SetVerbosity(int v) { verbosity_ = v; }
+
+   /// Set the target absolute residual norm tolerance for convergence
+   void SetTolerance(real_t tol) { const_tol_ = tol; }
+
+   /// Set the minimum number of nonzeros required for the solution.
+   void SetMinNNZ(int min_nnz) { min_nnz_ = min_nnz; }
+
+   /** @brief Set the maximum number of nonzeros required for the solution, as
+       an early termination condition. */
+   void SetMaxNNZ(int max_nnz) { max_nnz_ = max_nnz; }
+
+   /** @brief Set threshold on relative change in residual over nStallCheck_
+       iterations. */
+   void SetResidualChangeTolerance(real_t tol)
+   { res_change_termination_tol_ = tol; }
+
+   /** @brief Set the magnitude of projected residual entries that are
+       considered zero.  Increasing this value relaxes solution constraints. */
+   void SetZeroTolerance(real_t tol) { zero_tol_ = tol; }
+
+   /// Set RHS vector constant shift, defining rhs_lb and rhs_ub in Solve().
+   void SetRHSDelta(real_t d) { rhs_delta_ = d; }
+
+   /// Set the maximum number of outer iterations in Solve().
+   void SetOuterIterations(int n) { n_outer_ = n; }
+
+   /// Set the maximum number of inner iterations in Solve().
+   void SetInnerIterations(int n) { n_inner_ = n; }
+
+   /// Set the number of iterations to use for stall checking.
+   void SetStallCheck(int n) { nStallCheck_ = n; }
+
+   /// Set a flag to determine whether to call NormalizeConstraints().
+   void SetNormalize(bool n) { normalize_ = n; }
+
+   /** @brief
+     * Enumerated types of QRresidual mode. Options are 'off': the residual is
+     * calculated normally, 'on': the residual is calculated using the QR
+     * method, 'hybrid': the residual is calculated normally until we experience
+     * rounding errors, then the QR method is used. The default is 'hybrid',
+     * which should see the best performance. Recommend using 'hybrid' or 'off'
+     * only, since 'on' is computationally expensive.
+     */
+   enum class QRresidualMode {off, on, hybrid};
+
+   /** @brief
+    * Set the residual calculation mode for the NNLS solver. See QRresidualMode
+    * enum above for details.
+    */
+   void SetQRResidualMode(const QRresidualMode qr_residual_mode);
+
+   /**
+    * @brief Solve the NNLS problem. Specifically, we find a vector @a soln,
+    * such that rhs_lb < mat*soln < rhs_ub is satisfied, where mat is the
+    * DenseMatrix input to SetOperator().
+    *
+    * The method by which we find the solution is the active-set method
+    * developed by Lawson and Hanson (1974) using lapack. To decrease rounding
+    * errors in the case of very tight tolerances, we have the option to compute
+    * the residual using the QR factorization of A, by res = b - Q*Q^T*b. This
+    * residual calculation results in less rounding error, but is more
+    * computationally expensive. To select whether to use the QR residual method
+    * or not, see set_qrresidual_mode above.
+    */
+   void Solve(const Vector& rhs_lb, const Vector& rhs_ub, Vector& soln) const;
+
+   /** @brief
+     * Normalize the constraints such that the tolerances for each constraint
+     * (i.e. (UB - LB)/2) are equal. This seems to help the performance in most
+     * cases.
+     */
+   void NormalizeConstraints(Vector& rhs_lb, Vector& rhs_ub) const;
+
+private:
+   const DenseMatrix *mat;
+
+   real_t const_tol_;
+   int min_nnz_; // minimum number of nonzero entries
+   mutable int max_nnz_; // maximum number of nonzero entries
+   int verbosity_;
+
+   /**
+    * @brief Threshold on relative change in residual over nStallCheck_
+    * iterations, for stall sensing.
+    */
+   real_t res_change_termination_tol_;
+
+   real_t zero_tol_;
+   real_t rhs_delta_;
+   int n_outer_;
+   int n_inner_;
+   int nStallCheck_;
+
+   bool normalize_;
+
+   mutable bool NNLS_qrres_on_;
+   QRresidualMode qr_residual_mode_;
+
+   mutable Vector row_scaling_;
+};
+#endif // MFEM_USE_LAPACK
+
 }
 
 #endif // MFEM_SOLVERS
